@@ -551,6 +551,8 @@ class _TorSocksProtocol(Protocol):
 class _TorSocksFactory(Factory):
     protocol = _TorSocksProtocol
 
+    # XXX should do validation on this stuff so we get errors before
+    # building the protocol
     def __init__(self, *args, **kw):
         self._args = args
         self._kw = kw
@@ -689,6 +691,7 @@ class TorSocksEndpoint(object):
     # IAddress-implementer?
     def __init__(self, socks_endpoint, host, port, tls=False):
         self._proxy_ep = socks_endpoint  # can be Deferred
+        assert self._proxy_ep is not None
         if six.PY2 and isinstance(host, str):
             host = unicode(host)  # noqa
         if six.PY3 and isinstance(host, bytes):
@@ -716,7 +719,10 @@ class TorSocksEndpoint(object):
         if self._tls:
             # XXX requires Twisted 14+
             from twisted.internet.ssl import optionsForClientTLS
-            context = optionsForClientTLS(self._host)
+            if self._tls is True:
+                context = optionsForClientTLS(self._host)
+            else:
+                context = self._tls
             tls_factory = tls.TLSMemoryBIOFactory(context, True, factory)
             socks_factory = _TorSocksFactory(
                 self._host, self._port, 'CONNECT', tls_factory,
@@ -732,6 +738,11 @@ class TorSocksEndpoint(object):
         # XXX isn't this just maybeDeferred()
         if isinstance(self._proxy_ep, Deferred):
             proxy_ep = yield self._proxy_ep
+            if not IStreamClientEndpoint.providedBy(proxy_ep):
+                raise ValueError(
+                    "The Deferred provided as 'socks_endpoint' must "
+                    "resolve to an IStreamClientEndpoint provider"
+                )
         else:
             proxy_ep = self._proxy_ep
 
