@@ -7,7 +7,6 @@ from __future__ import with_statement
 import collections
 import os
 import stat
-import types
 import warnings
 
 from twisted.internet import defer
@@ -127,11 +126,13 @@ def build_local_tor_connection(reactor, host='127.0.0.1', port=9051,
                                socket='/var/run/tor/control', *args, **kwargs):
     """
     This builds a connection to a local Tor, either via 127.0.0.1:9051
-    (which is tried first) or /var/run/tor/control (by default). See
-    also :meth:`build_tor_connection
+    or /var/run/tor/control (by default; the latter is tried
+    first). See also :meth:`build_tor_connection
     <txtorcon.torstate.build_tor_connection>` for other key-word
     arguments that are accepted here also.
 
+    **Note**: new code should use :meth:`txtorcon.connect` instead.
+
     :param host:
         An IP address to find Tor at. Corresponds to the
         ControlListenAddress torrc option.
@@ -144,7 +145,7 @@ def build_local_tor_connection(reactor, host='127.0.0.1', port=9051,
 
     try:
         return build_tor_connection((reactor, socket), *args, **kwargs)
-    except:
+    except Exception:
         return build_tor_connection((reactor, host, port), *args, **kwargs)
 
 
@@ -210,8 +211,9 @@ class TorState(object):
     :class:`txtorcon.interface.IRouterContainer`.
 
     :cvar DO_NOT_ATTACH:
-    Constant to return from an IAttacher indicating you don't want to
-    attach this stream at all.
+        Constant to return from an IAttacher indicating you don't want to
+        attach this stream at all.
+
     """
 
     @classmethod
@@ -256,6 +258,7 @@ class TorState(object):
 
         #: keys by hexid (string) and by unique names
         self.routers = {}
+        self._old_routers = {}
 
         #: keys on name, value always list (many duplicate "Unnamed"
         #: routers, for example)
@@ -290,10 +293,11 @@ class TorState(object):
     def _create_router(self, **kw):
         id_hex = hexIdFromHash(kw['idhash'])
         try:
-            router = self.routers[id_hex]
+            router = self._old_routers[id_hex]
         except KeyError:
             router = Router(self.protocol)
-            self.routers[router.id_hex] = router
+
+        self.routers[id_hex] = router
         router.from_consensus = True
         router.update(
             kw['nickname'],
@@ -315,15 +319,11 @@ class TorState(object):
         if 'authority' in router.flags:
             self.authorities[router.name] = router
 
-        if router.id_hex in self.routers:
-            # FIXME should I do an update() on this one??
-            router = self.routers[router.id_hex]
-        else:
-            if router.name in self.routers:
-                self.routers[router.name] = None
+        if router.name in self.routers:
+            self.routers[router.name] = None
 
-            else:
-                self.routers[router.name] = router
+        else:
+            self.routers[router.name] = router
 
         if router.name in self.routers_by_name:
             self.routers_by_name[router.name].append(router)
@@ -349,6 +349,7 @@ class TorState(object):
             'ns/all',
             self._network_status_parser.feed_line,
         )
+        self._network_status_parser.done()
 
         # update list of existing circuits
         cs = yield self.protocol.get_info_raw('circuit-status')
@@ -368,7 +369,7 @@ class TorState(object):
                 continue            # FIXME
             self.addrmap.update(line)
 
-        self._add_events()
+        yield self._add_events()
 
         entries = yield self.protocol.get_info_raw("entry-guards")
         for line in entries.split('\n')[1:]:
@@ -404,7 +405,7 @@ class TorState(object):
             try:
                 pid = parse_keywords(pid)['process/pid']
                 self.tor_pid = int(pid)
-            except:
+            except Exception:  # fixme: ValueError and KeyError ..?
                 self.tor_pid = 0
         if not self.tor_pid and self.protocol.is_owned:
             self.tor_pid = self.protocol.is_owned
@@ -412,6 +413,7 @@ class TorState(object):
         self.post_bootstrap.callback(self)
         self.post_boostrap = None
 
+    # XXX this should be hidden as _undo_attacher
     def undo_attacher(self):
         """
         Shouldn't Tor handle this by turning this back to 0 if the
@@ -533,12 +535,20 @@ class TorState(object):
         )
 
     def add_circuit_listener(self, icircuitlistener):
+        """
+        Adds a new instance of :class:`txtorcon.interface.ICircuitListener` which
+        will receive updates for all existing and new circuits.
+        """
         listen = ICircuitListener(icircuitlistener)
         for circ in self.circuits.values():
             circ.listen(listen)
         self.circuit_listeners.append(listen)
 
     def add_stream_listener(self, istreamlistener):
+        """
+        Adds a new instance of :class:`txtorcon.interface.IStreamListener` which
+        will receive updates for all existing and new streams.
+        """
         listen = IStreamListener(istreamlistener)
         for stream in self.streams.values():
             stream.listen(listen)
@@ -600,6 +610,7 @@ class TorState(object):
 
     DO_NOT_ATTACH = object()
 
+    # @defer.inlineCallbacks  (this method is async, be nice to mark it ...)
     def _maybe_attach(self, stream):
         """
         If we've got a custom stream-attachment instance (see
@@ -642,7 +653,9 @@ class TorState(object):
             txtorlog.msg("circuit:", circ)
             if circ is None or circ is TorState.DO_NOT_ATTACH:
                 # tell Tor to do what it likes
-                return self.protocol.queue_command(b"ATTACHSTREAM %d 0" % stream.id)
+                return self.protocol.queue_command(
+                    u"ATTACHSTREAM {} 0".format(stream.id).encode("ascii")
+                )
 
             else:
                 # should get a Circuit instance; check it for suitability
@@ -662,7 +675,7 @@ class TorState(object):
                     )
                 # we've got a valid Circuit instance; issue the command
                 return self.protocol.queue_command(
-                    b"ATTACHSTREAM %d %d" % (stream.id, circ.id)
+                    u"ATTACHSTREAM {} {}".format(stream.id, circ.id).encode("ascii")
                 )
 
         circ_d.addCallback(issue_stream_attach)
@@ -717,9 +730,13 @@ class TorState(object):
         from NEWCONSENSUS events.
         """
 
-        # XXX why are we getting this with 0 data?
+        # XXX why are we ever getting this with 0 data?
         if len(data):
+            self._old_routers = self.routers
+            self.routers = dict()
             self.all_routers = set()
+            self.routers_by_hash = dict()
+            self.routers_by_name = dict()
             for line in data.split('\n'):
                 self._network_status_parser.feed_line(line)
             self._network_status_parser.done()
@@ -731,7 +748,6 @@ class TorState(object):
             if v is None:
                 txtorlog.msg(len(self.routers_by_name[k]), "dups:", k)
                 remove_keys.add(k)
-
         for k in remove_keys:
             del self.routers[k]
 
@@ -796,28 +812,23 @@ class TorState(object):
         txtorlog.msg(" --> addr_map", addr)
         self.addrmap.update(addr)
 
-    event_map = {'STREAM': _stream_update,
-                 'CIRC': _circuit_update,
-                 'NEWCONSENSUS': _update_network_status,
-                 'ADDRMAP': _addr_map}
-    """event_map used by add_events to map event_name -> unbound method"""
+    event_map = {
+        'STREAM': '_stream_update',
+        'CIRC': '_circuit_update',
+        'NEWCONSENSUS': '_update_network_status',
+        'ADDRMAP': '_addr_map',
+    }
+
     @defer.inlineCallbacks
     def _add_events(self):
         """
         Add listeners for all the events the controller is interested in.
         """
 
-        for (event, func) in self.event_map.items():
-            # the map contains unbound methods, so we bind them
-            # to self so they call the right thing
-            try:
-                bound = types.MethodType(func, self, TorState)
-            except TypeError:
-                # python3
-                bound = types.MethodType(func, self)
+        for (event, func_name) in self.event_map.items():
             yield self.protocol.add_event_listener(
                 event,
-                bound,
+                getattr(self, func_name),
             )
 
     # ICircuitContainer
